@@ -44,6 +44,14 @@
 #endif
 #include <gdk/gdkkeysyms.h>
 
+//xml
+#include <stdio.h>
+#include <stdlib.h>
+#include <libxml/xmlmemory.h>
+#include <libxml/parser.h>
+
+#define LINE_LIMIT 500
+
 /* Privates: */
 void safeReplaceString(char** oldstr, const gchar *newstr, MainView *mainview);
 void maybe_save_file(MainView *mainview);
@@ -424,7 +432,7 @@ void callback_file_open(GtkAction * action, MainView *mainview) {
 	g_assert(mainview != NULL && mainview->data != NULL );
  	/* maybe_save_file(mainview); */
     save_mod_file(mainview);
-	mainview->file_name = interface_file_chooser ( mainview, GTK_FILE_CHOOSER_ACTION_OPEN );
+	mainview->file_name = interface_file_chooser ( mainview, GTK_FILE_CHOOSER_ACTION_OPEN);
 	read_pwsafe_file(mainview);
 }
 
@@ -644,12 +652,13 @@ void callback_pwdialog_button(GtkWidget *widget, PasswordUI *pwUi) {
 
 /* export */
 void callback_export ( GtkAction * action, MainView *mainview ) {
-    char *fname;
+    gchar *fname;
     
     fname = interface_file_chooser(mainview, GTK_FILE_CHOOSER_ACTION_SAVE);
     if (fname != NULL) {
         FILE *outf;
         outf = fopen(fname,"w");
+        g_free(fname);
         if (outf != (FILE *) NULL) {
             GList *onrec=NULL;
      
@@ -672,6 +681,272 @@ void callback_export ( GtkAction * action, MainView *mainview ) {
     }
 }
 
+gint importXmlKeePassX(MainView *mainview, xmlDocPtr doc)  {
+	gint imported = 0;
+	xmlNodePtr cur;
+	xmlNodePtr curGroup;
+	xmlNodePtr curEntry;
+	xmlChar *groupName = NULL;
+	xmlChar *entryName = NULL;
+	xmlChar *userName = NULL;
+	xmlChar *password = NULL;
+	xmlChar *notes2 = NULL;
+	xmlChar *notes = NULL;
+	xmlChar *notes1 = NULL;
+	int strLen = 0;
+	
+	rec * record;
+
+	cur = xmlDocGetRootElement(doc);
+
+	if (cur == NULL) {
+		fprintf(stderr,"empty document\n");
+	} else {
+		if (xmlStrcasecmp(cur->name, (const xmlChar *) "database")) {
+			msgbox(mainview,_("document of the wrong type, root node != database"));
+		} else {
+			//Group
+			curGroup = cur->xmlChildrenNode;
+			while (curGroup != NULL) {
+				if ((!xmlStrcasecmp(curGroup->name, (const xmlChar *)"group"))){
+					cur = curGroup->xmlChildrenNode;
+					while (cur != NULL) {
+						if ((!xmlStrcasecmp(cur->name, (const xmlChar *)"title"))){
+							if (groupName != NULL) {
+								xmlFree(groupName);
+							}
+							groupName = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
+						} else if ((!xmlStrcasecmp(cur->name, (const xmlChar *)"entry"))){
+							// loop on entry
+							curEntry = cur->xmlChildrenNode;
+							while (curEntry != NULL) {
+								if ((!xmlStrcasecmp(curEntry->name, (const xmlChar *)"title"))){
+									entryName = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+								} else if ((!xmlStrcasecmp(curEntry->name, (const xmlChar *)"username"))){
+									userName = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+								} else if ((!xmlStrcasecmp(curEntry->name, (const xmlChar *)"password"))){
+									password = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+								} else if ((!xmlStrcasecmp(curEntry->name, (const xmlChar *)"url")) || (!xmlStrcasecmp(curEntry->name, (const xmlChar *)"comment"))){
+									if (notes != NULL) {
+										notes2 = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+										if (notes2 != NULL) {
+											strLen = xmlStrlen(notes2);
+											if (strLen > LINE_LIMIT) {
+												strLen = LINE_LIMIT;
+											}
+											if (strLen > 0) {
+												notes1 = xmlStrncatNew(notes, "\r\n", 2);
+												xmlFree(notes);
+												notes = xmlStrncatNew(notes1, notes2, strLen);
+												xmlFree(notes1);
+												xmlFree(notes2);
+											}
+										}
+									} else {
+										notes = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+										if ((notes != NULL) && (xmlStrlen(notes) == 0)) {
+											xmlFree(notes);
+											notes = NULL;
+										}
+									}
+								}
+								curEntry = curEntry->next;
+							}
+							if (entryName != NULL || userName != NULL) {
+								if (groupName == NULL) {
+									groupName = xmlCharStrdup(_("Group"));
+								}
+								imported ++;
+								record = add_new_record((entryName == NULL)?userName:entryName, groupName);
+								record->user = g_strdup(userName);
+								record->password = g_strdup(password);
+								record->notes = g_strdup(notes);
+							}
+							if (entryName != NULL) {
+								xmlFree(entryName);
+								entryName = NULL;
+							}
+							if (userName != NULL) {
+								xmlFree(userName);
+								userName = NULL;
+							}
+							if (password != NULL) {
+								xmlFree(password);
+								password = NULL;
+							}
+							if (notes != NULL) {
+								xmlFree(notes);
+								notes = NULL;
+							}
+						}
+						cur = cur->next;
+					}
+				}
+				curGroup = curGroup->next;
+			}
+			if (groupName != NULL) {
+				xmlFree(groupName);
+				groupName = NULL;
+			}
+		}
+	}
+	return imported;
+}
+
+gint importXmlFPM2(MainView *mainview, xmlDocPtr doc)  {
+	gint imported = 0;
+	xmlNodePtr cur;
+	xmlNodePtr curGroup;
+	xmlNodePtr curEntry;
+	xmlChar *groupName = NULL;
+	xmlChar *entryName = NULL;
+	xmlChar *userName = NULL;
+	xmlChar *password = NULL;
+	xmlChar *notes2 = NULL;
+	xmlChar *notes = NULL;
+	xmlChar *notes1 = NULL;
+	int strLen = 0;
+	
+	rec * record;
+
+	cur = xmlDocGetRootElement(doc);
+
+	if (cur == NULL) {
+		fprintf(stderr,"empty document\n");
+	} else {
+		if (xmlStrcasecmp(cur->name, (const xmlChar *) "FPM")) {
+			msgbox(mainview,_("document of the wrong type, root node != FPM"));
+		} else {
+			//List
+			cur = cur->xmlChildrenNode;
+			while (cur != NULL) {
+				if ((!xmlStrcasecmp(cur->name, (const xmlChar *)"PasswordList"))){
+					cur = cur->xmlChildrenNode;
+					while (cur != NULL) {
+						if ((!xmlStrcasecmp(cur->name, (const xmlChar *)"PasswordItem"))){
+							// loop on entry
+							curEntry = cur->xmlChildrenNode;
+							while (curEntry != NULL) {
+								if ((!xmlStrcasecmp(curEntry->name, (const xmlChar *)"title"))){
+									entryName = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+								} else if ((!xmlStrcasecmp(curEntry->name, (const xmlChar *)"user"))){
+									userName = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+								} else if ((!xmlStrcasecmp(curEntry->name, (const xmlChar *)"password"))){
+									password = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+								} else if ((!xmlStrcasecmp(curEntry->name, (const xmlChar *)"category"))){
+									groupName = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+								} else if ((!xmlStrcasecmp(curEntry->name, (const xmlChar *)"url")) || (!xmlStrcasecmp(curEntry->name, (const xmlChar *)"notes"))){
+									if (notes != NULL) {
+										notes2 = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+										if (notes2 != NULL) {
+											strLen = xmlStrlen(notes2);
+											if (strLen > LINE_LIMIT) {
+												strLen = LINE_LIMIT;
+											}
+											if (strLen > 0) {
+												notes1 = xmlStrncatNew(notes, "\r\n", 2);
+												xmlFree(notes);
+												notes = xmlStrncatNew(notes1, notes2, strLen);
+												xmlFree(notes1);
+												xmlFree(notes2);
+											}
+										}
+									} else {
+										notes = xmlNodeListGetString(doc, curEntry->xmlChildrenNode, 1);
+										if ((notes != NULL) && (xmlStrlen(notes) == 0)) {
+											xmlFree(notes);
+											notes = NULL;
+										}
+									}
+								}
+								curEntry = curEntry->next;
+							}
+							if (entryName != NULL || userName != NULL) {
+								if (groupName == NULL) {
+									groupName = xmlCharStrdup(_("Group"));
+								}
+								imported ++;
+								record = add_new_record((entryName == NULL)?userName:entryName, groupName);
+								record->user = g_strdup(userName);
+								record->password = g_strdup(password);
+								record->notes = g_strdup(notes);
+							}
+							xmlFree(groupName);
+							if (entryName != NULL) {
+								xmlFree(entryName);
+								entryName = NULL;
+							}
+							if (userName != NULL) {
+								xmlFree(userName);
+								userName = NULL;
+							}
+							if (password != NULL) {
+								xmlFree(password);
+								password = NULL;
+							}
+							if (notes != NULL) {
+								xmlFree(notes);
+								notes = NULL;
+							}
+						}
+						cur = cur->next;
+					}
+				}
+				if (cur != NULL) {
+					cur = cur->next;
+				}
+			}
+		}
+	}
+	return imported;
+}
+
+void callback_import_xml(GtkAction * action, MainView *mainview ) {
+	xmlDocPtr doc;
+	gchar *notes = NULL;
+	gchar *fname;
+	
+	gint imported = 0;
+	gint xmlFormat = PWSAFE_KEEPASSX;
+	
+	xmlFormat = interface_getXmlFormat(mainview, xmlFormat);
+	if (xmlFormat < 0) {
+		return;
+	}
+	if (xmlFormat >= PWSAFE_NB_XML_IMPORT) {
+		msgbox(mainview, _("Import XML Format not yet supported!"));
+		return;
+	}
+	
+    fname = interface_file_chooser(mainview, GTK_FILE_CHOOSER_ACTION_OPEN);
+    if (fname != NULL) {
+		doc = xmlParseFile(fname);
+	
+		if (doc == NULL ) {
+			msgbox(mainview,_("Document not parsed successfully."));
+		} else {
+			if (xmlFormat == PWSAFE_KEEPASSX) {
+				imported = importXmlKeePassX(mainview, doc);
+			} else if (xmlFormat == PWSAFE_FPM2) {
+				imported = importXmlFPM2(mainview, doc);
+			} else {
+				msgbox(mainview,_("Import XML Format not yet supported!"));
+			}
+			xmlFreeDoc(doc);
+		}
+		g_free(fname);
+		if (imported > 0) {
+			populate_treeView(mainview);
+			notes = g_strdup_printf(_("%d items imported."), imported);
+			msgbox(mainview, notes);
+			g_free(notes);
+		} else {
+			msgbox(mainview, _("No item imported."));
+		}
+	}
+}
+
+	
 DBusHandlerResult dbus_handle_mce_message(DBusConnection *con, DBusMessage *msg, MainView *mainView ) {
 	DBusMessageIter iter;
 	dbus_bool_t value;
